import os, shutil
import subprocess
import hashlib
import threading
import memigui as g
from memigui import *
from mpkgenums import CanMsgType, EMMsgType, TouchEvent, EMNotifyType, EMEncEvent

# import inspect
# from pprint import *

# Main class which will be called from package manager
class pkgrunner():
  
  # Reference to the calling mpkg manager class 
  mpkgmanager=None
  msgBoxStd = None
  pkgTitleText = "Vibes Data Management"

  def __init__(self, mgr) -> None:
    self.mpkgmanager = mgr
    mpkgmanager = self.mpkgmanager

    self.functions = {
      "CanID_90": self.OnEventEnc1, 
      "CanID_91": self.OnEventEnc2, 
      "CanID_92": self.OnEventEnc3, 
      "CanID_93": self.OnEventEnc4, 
      "CanID_90_0": self.NotImplemented, 
      "CanID_91_0": self.NotImplemented, 
    }

    self.rsyncScope = {
      0: "",
      1: "/presets",
      2: "/wave",
      3: "/wave",
      4: "/wave/noises",
      5: "/wave/drumkit",
      6: "/midifiles"
    }

    self.rsyncArgs = {
      0: "-r",
      1: "-r",
      2: "",
      3: "-r",
      4: "-r",
      5: "-r",
      6: "-r"
    }

    self.states = {
      "facreset" : "",
      "restore" : ""
    }

    # -------------- Define Styles to be used beyond defaults -------------- 
    
    g.CTheme.SetNamedColor("MPMMsgBoxBorder",g.CTheme.ColorFromRGB(50,50,50))

    g.CTheme.SetNamedColor("BtnFg",g.CTheme.ColorFromRGB(0,0,255))
    g.CTheme.SetNamedColor("BtnFgAcc",g.CTheme.ColorFromRGB(16,16,64))
    g.CTheme.SetNamedColor("BtnBorder",g.CTheme.ColorFromRGB(255,0,0))
    g.CTheme.SetNamedColor("BtnBorderAcc",g.CTheme.ColorFromRGB(0,0,255))
    g.CTheme.SetNamedColor("BtnText",g.CTheme.ColorFromRGB(255,255,255))
    g.CTheme.SetNamedColor("BtnTextAcc",g.CTheme.ColorFromRGB(48,48,48))
    g.CTheme.SetNamedColor("ValLblBg", g.CTheme.ColorFromRGB(32, 32, 32))
    #g.CTheme.SetNamedColor("ValLblBorderAcc", g.CTheme.ColorFromRGB(192, 0, 0))
    g.CTheme.SetNamedColor("ValLblBorderAcc", g.CTheme.ColorFromRGB(90, 156, 162))
    g.CTheme.SetNamedColor("ValLblTitleText", g.CTheme.ColorFromRGB(164, 164, 164))
    g.CTheme.SetNamedColor("ValLblText", g.CTheme.ColorFromRGB(192, 192, 192))
    g.CTheme.SetNamedColor("ValLblTextAcc", g.CTheme.ColorFromRGB(255, 255, 255))

    # -------------- Screen play --------------
    if os.path.exists("bg.png"):
      self.bgimage=g.emImage(mpkgmanager.gui, emPOINT(0, 0), "bg.png", "./")
      bgimage=self.bgimage
      mpkgmanager.gui.AddElement(bgimage)
      bgimage.OnPaint()

    
    if os.path.exists("pkg.png"):
      self.rtimage=g.emImage(mpkgmanager.gui, emPOINT(530, 30), "pkg.png", "./")
      rtimage=self.rtimage
      mpkgmanager.gui.AddElement(rtimage)
      rtimage.OnPaint()

    self.msgBoxStd=g.emMessageBox(mpkgmanager.gui,"MPM",500,305,0x0,0x00ffffff,0x00ffffff)
    msgBoxStd = self.msgBoxStd
    msgBoxStd.show(emPOINT(10,10), self.pkgTitleText)
    msgBoxStd.OnPaint()
    msgBoxStd.setText("From Init")
    #self.LogMessage(f"Called with Package path {mpkgmanager.pkgPath}")

    self.textSpinner = g.emLabel(mpkgmanager.gui, "Spinner", emPOINT(400, 330), 60, 60, g.CTheme.ColorFromRGB(34,34,34), g.CTheme.ColorFromRGB(34,34,34), g.CTheme.ColorFromRGB(255,255,255), 14)
    textSpinner = self.textSpinner
    textSpinner.setText("")
    textSpinner.OnPaint()
    mpkgmanager.gui.AddElement(textSpinner)
    self.runSpinner = False

    self.btnW1=emButton(mpkgmanager.gui, 1001, emPOINT(10, 405), 180, 60, g.CTheme.ColorFromRGB(16,16,64), g.CTheme.ColorFromRGB(0,0,255), g.CTheme.ColorFromRGB(255,255,255), True, 1, 0)
    btnW1=self.btnW1
    btnW1.SetText("Backup")
    btnW1.setRepaint()
    btnW1.OnPaint()
    btnW1.SetFocus(True)
    mpkgmanager.gui.AddElement(btnW1)

    self.btnW2=emButton(mpkgmanager.gui, 1002, emPOINT(210, 405), 180, 60, g.CTheme.ColorFromRGB(16,16,64), g.CTheme.ColorFromRGB(0,0,255), g.CTheme.ColorFromRGB(255,255,255), True, 1, 0)
    btnW2=self.btnW2
    btnW2.SetText("Restore")
    btnW2.setRepaint()
    btnW2.OnPaint()
    btnW2.SetFocus(True)
    mpkgmanager.gui.AddElement(btnW2)

    self.btnW3=None
    # self.btnW3=emButton(mpkgmanager.gui, 1003, emPOINT(410, 405), 180, 60, g.CTheme.ColorFromRGB(16,16,64), g.CTheme.ColorFromRGB(0,0,255), g.CTheme.ColorFromRGB(255,255,255), True, 1, 0)
    # btnW3=self.btnW3
    # btnW3.SetText("Reset")
    # btnW3.setRepaint()
    # btnW3.OnPaint()
    # btnW3.SetFocus(True)
    # mpkgmanager.gui.AddElement(btnW3)

    # self.btnW4=emButton(mpkgmanager.gui, 1004, emPOINT(610, 330), 180, 60, g.CTheme.ColorFromRGB(16,16,64), g.CTheme.ColorFromRGB(0,0,255), g.CTheme.ColorFromRGB(255,255,255), True, 1, 0)
    # btnW4=self.btnW4
    # btnW4.SetText("BTN 4")
    # btnW4.setRepaint()
    # btnW4.OnPaint()
    # btnW4.SetFocus(True)
    # mpkgmanager.gui.AddElement(btnW4)

    self.vlblW2=emValueLabel(mpkgmanager.gui, "ActSel", 2001, 0, emPOINT(10, 330), 380, 60, g.CTheme.ColorFromName("ValLblBg"), g.CTheme.ColorFromName("ValLblBorderAcc"), g.CTheme.ColorFromName("ValLblText"), g.CTheme.ColorFromName("ValLblTextAcc"), g.CTheme.ColorFromName("ValLblTitleText"))
    vlblW2 = self.vlblW2
    vlblW2.SetLabelText(" Backup/Restore ", "", 12)
    vlblW2.AddEnumText("All")
    vlblW2.AddEnumText("Presets")
    vlblW2.AddEnumText("Wavetables")
    vlblW2.AddEnumText("All wave files")
    vlblW2.AddEnumText("Noise wave files")
    vlblW2.AddEnumText("Drumkit wave files")
    vlblW2.AddEnumText("Midi files")
    vlblW2.SetFloatValue(0.0, 0)
    vlblW2.SetEncoderMode(8)
    vlblW2.SetChecked(True)
    vlblW2.SetFocus(True)
    vlblW2.SetEnabled(True)
    mpkgmanager.gui.AddElement(vlblW2)

    self.btnExit=emButton(mpkgmanager.gui, 9004, emPOINT(610, 405), 180, 60, g.CTheme.ColorFromRGB(16,16,64), g.CTheme.ColorFromRGB(0,0,255), g.CTheme.ColorFromRGB(255,255,255), True, 1, 0)
    btnExit=self.btnExit
    btnExit.SetText("Exit")
    btnExit.setRepaint()
    btnExit.OnPaint()
    btnExit.SetFocus(True)
    mpkgmanager.gui.AddElement(btnExit)

  def spinnerCtl(self, state):
    import time
    # Function to show progress
    def spinnerrun():
        spinner_chars = ['|', '/', '-', '\\']  # Spinner characters
        index = 0  # Index to track the current character
        while self.runSpinner:
            time.sleep(0.2)  # Wait for 100ms
            self.textSpinner.setText(spinner_chars[index])  # Set spinner character
            self.textSpinner.OnPaint()
            index = (index + 1) % len(spinner_chars)  # Update index for next character

    # Start the progress display in a separate thread
    import threading
    if state and not self.runSpinner:
      self.runSpinner = True
      self.progress_thread = threading.Thread(target=spinnerrun)
      self.progress_thread.start()
    elif not state and self.runSpinner:
      self.runSpinner = False
      self.progress_thread.join()
      self.textSpinner.setText("")
      self.textSpinner.OnPaint()

  def run(self):
    dashes = "-----------------------------------------------"
    dbefore = (45 - len(self.pkgTitleText))/2
    dafter = int(dbefore)
    if dbefore % 1 > 0:
      dafter+=1
    dbefore = int(dbefore)

    #self.LogMessage("Run function of pkgrunner called from mpkg manager object")
    #self.LogMessage("--------- Vibes Update V3.03 Beta 1 -----------")
    self.LogMessage(f"{dashes[:dbefore]} {self.pkgTitleText} {dashes[-dafter:]}")
    self.LogMessage(" ")
    self.LogMessage("Press Wheel 1 to run Backup ")
    self.LogMessage("Press Wheel 2 to run Restore ")
    self.LogMessage("Turn  Wheel 1 or 2 to filter items")
    self.LogMessage("Press Wheel 3 to run Factory Reset ")
    self.LogMessage("Press Wheel 4 to Exit ")
    self.LogMessage(" ")
    self.LogMessage("-----------------------------------------------")
    self.LogMessage(" ")


  def OnGuiMessage(self, msg):

    try:
      em_msg_t = EMMsgType(msg.type)
    except:
      em_msg_t = EMMsgType.UNKNOWN

    try:
      touch_event = TouchEvent(msg.m_evt_type)
    except:
      touch_event = TouchEvent.UNKNOWN

    # print('MPKG Message received: msg-type:', em_msg_t, " event:", touch_event, " pData:", msg.pData, " arg:", msg.arg, " flag:", msg.flag, " param:", msg.param, " x,y:", msg.m_x, ",", msg.m_y)

    if em_msg_t == EMMsgType.EM_MSG_ON_BUTTON:
        self.BtnOnEventHandler(em_msg_t, msg.param)
  
  def LogMessage(self, text, single=False):
    msgBoxStd = self.msgBoxStd
    chunk_size = 47

    if single:
      msgBoxStd.setText(text)
      msgBoxStd.setRepaint()
      msgBoxStd.OnPaint()
    else:
      msgBoxStd.setText("")
      text_list = text.split("\n")

      for t in text_list:
        for i in range(0, len(t), chunk_size):
          chunk = t[i:i+chunk_size]
          msgBoxStd.AddLine(chunk)

  def ShowLogWindow(self, visibility):
    msgBoxStd = self.msgBoxStd
    mgr = self.mpkgmanager
    if visibility:
      msgBoxStd.show(emPOINT(10,10), self.pkgTitleText)
      msgBoxStd.setRepaint()
      msgBoxStd.OnPaint()
    else:
      msgBoxStd.hide()
      msgBoxStdRect = msgBoxStd.getClientRect()
      mgr.dc.BeginPaint(0, 0, 800, 480, False, False)
      mgr.dc.FillFrame(msgBoxStdRect.left, msgBoxStdRect.top, msgBoxStdRect.right, msgBoxStdRect.bottom ,g.CTheme.ColorFromRGB(0,0,0))
      mgr.dc.EndPaint()

  def OnCanMessage(self, id, byte_str, len):
    #print(f"MPKG Can Message received: id={id}, data={byte_str}, len={len}")

    try:
      f = f'CanID_{hex(id)[2:]}'
      self.functions[f](byte_str, len)
    except KeyError:
      print(f"No Function mapping for '{f}' found")
  
  #def OnEventEnc1(self, data, len):
  #  try:
  #    enc_evt = EMEncEvent(data[0])
  #  except:
  #    enc_evt = EMEncEvent.UNKNOWN
  #  
  #  if enc_evt == EMEncEvent.PRESS:
  #    # self.LogMessage(f'F: {self.vlblW2.GetFloatValue()} I: {self.vlblW2.GetIntValue()}')
  #    try:
  #      f = f'CanID_90_{int(self.vlblW2.GetFloatValue())}'
  #      self.functions[f]()
  #    except KeyError:
  #      self.LogMessage(f"No Function mapping for '{f}' found")
  #
  #  else:
  #    msg=g.tEM_Message()
  #    msg.type=EMMsgType.EM_MSG_ENCODER_EVENT.value
  #    msg.flag=data[0]
  #    msg.param=0
  #    self.vlblW2.OnMessage(msg)

  def OnEventEnc1(self, data, len):
    try:
      enc_evt = EMEncEvent(data[0])
    except:
      enc_evt = EMEncEvent.UNKNOWN
    
    if enc_evt == EMEncEvent.PRESS:
      # self.LogMessage(f'F: {self.vlblW2.GetFloatValue()} I: {self.vlblW2.GetIntValue()}')
      self.runBackup()
    else:
      msg=g.tEM_Message()
      msg.type=EMMsgType.EM_MSG_ENCODER_EVENT.value
      msg.flag=data[0]
      msg.param=0
      self.vlblW2.OnMessage(msg)

  def OnEventEnc2(self, data, len):
    try:
      enc_evt = EMEncEvent(data[0])
    except:
      enc_evt = EMEncEvent.UNKNOWN
    
    if enc_evt == EMEncEvent.PRESS:
      # self.LogMessage(f'F: {self.vlblW2.GetFloatValue()} I: {self.vlblW2.GetIntValue()}')
      self.runRestore()
    else:
      msg=g.tEM_Message()
      msg.type=EMMsgType.EM_MSG_ENCODER_EVENT.value
      msg.flag=data[0]
      msg.param=0
      self.vlblW2.OnMessage(msg)

  def OnEventEnc3(self, data, len):
    try:
      enc_evt = EMEncEvent(data[0])
    except:
      enc_evt = EMEncEvent.UNKNOWN
    
    if enc_evt == EMEncEvent.PRESS:
      # self.LogMessage(f'F: {self.vlblW2.GetFloatValue()} I: {self.vlblW2.GetIntValue()}')
      self.runReset()

  def OnEventEnc4(self, data, len):
    try:
      enc_evt = EMEncEvent(data[0])
    except:
      enc_evt = EMEncEvent.UNKNOWN
    
    if enc_evt == EMEncEvent.PRESS:
      # self.LogMessage(f'F: {self.vlblW2.GetFloatValue()} I: {self.vlblW2.GetIntValue()}')
      self.OnExit()

  def OnExit(self):
    self.spinnerCtl(False)
    self.LogMessage("Exited.")
    self.mpkgmanager.quit()

  def find_directory(self, findddirname, startdirname):
    """Find the full path of the 'vibes-backup' directory starting from the given directory."""
    for dirpath, dirnames, filenames in os.walk(startdirname):
      if findddirname in dirnames:
        return os.path.join(dirpath, findddirname)
    return None

  def check_files_hash(self, datadir, hashfile):
      """Check if the hashes of files in hashfile match their corresponding files."""

      # Read the expected hashes from the hash file
      expected_hashes = {}
      with open(hashfile, 'r') as hash_file:
          for line in hash_file:
              parts = line.split()
              if len(parts) == 2:  # Ensure there are two parts: hash and filename
                  expected_hashes[parts[1]] = parts[0]  # Map filename to its expected hash

      # Check each file's hash
      for filename, expected_hash in expected_hashes.items():
          data_file_path = os.path.join(datadir, filename)

          # Calculate the hash of the data file
          sha256_hash = hashlib.sha256()
          try:
              with open(data_file_path, 'rb') as data_file:
                  for byte_block in iter(lambda: data_file.read(4096), b""):
                      sha256_hash.update(byte_block)
              calculated_hash = sha256_hash.hexdigest()

              # Compare the hashes
              if calculated_hash != expected_hash:
                  self.LogMessage(f"ERROR: verify failed for {filename} ... stopping")
                  return False
              else:
                  self.LogMessage(f"... {filename} ... ok")                
          except FileNotFoundError:
              self.LogMessage(f"ERROR: File {filename} not found.")
              return False

      return True


  def runBackup(self):
    self.LogMessage(f"Run Backup action ...")
    backupdir = self.find_directory("vibes-backup", "/tmp/media")
    if not backupdir:
      self.LogMessage(f"Backupdirectory 'vibes-backup' not found ... stopping") 
      return
    
    idx = int(self.vlblW2.GetFloatValue())
    scope = self.rsyncScope[idx]
    args = self.rsyncArgs[idx]

    #cmd = f"rsync {args} -L --info=progress2 /var/lib/emi/working{scope}/* {backupdir}{scope}"
    cmd = f"rsync {args} -L --info=name1 /var/lib/emi/working{scope}/* {backupdir}{scope}"

    self.spinnerCtl(True)

    # Use the cmd variable as a single command string
    process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    #process = subprocess.Popen(['python3', 'flashtest.py'], stdout=subprocess.PIPE, bufsize=256, universal_newlines=True)
    self.LogMessage("Start backup ...")

    # Read output while the process is running
    while process.poll() is None:
      
      output = process.stdout.read1().decode().strip()
      #self.LogMessage(f"{output}", True)
      self.LogMessage(f"{output}")

    # Read remaining output after the process has finished
    #output = process.communicate()[0].decode().strip()
    #if output:
    #  self.LogMessage(output)

    self.spinnerCtl(False)
    self.LogMessage("Backup ... done")

  def runRestore(self):

    def confirm_timer():
      import time
      time.sleep(10)  # Wait for 10s
      if self.states["restore"] == "wait-to-confirm":
        self.LogMessage(f" Run Restore without delete ... ")
        self.states["restore"] = "wait-to-confirm-answer"
        self.runRestore()

    if self.states["restore"] == "":
      self.LogMessage(f" Delete items?") 
      self.LogMessage(f" (!) To delete items which don't exist in your") 
      self.LogMessage(f" (!) backup press wheelknob 2 or touch 'Yes'") 
      self.LogMessage(f" Do continue without delete, wait. ")
      self.states["restore"] = "wait-to-confirm"
      self.confirm_thread = threading.Thread(target=confirm_timer)
      self.confirm_thread.start()
      btnW2=self.btnW2
      btnW2.SetText("Yes")
      btnW2.setRepaint()
      btnW2.OnPaint()
      return
    
    doDelete = True
    if self.states["restore"] == "wait-to-confirm-answer":
      doDelete = False

    btnW2=self.btnW2
    btnW2.SetText("Restore")
    btnW2.setRepaint()
    btnW2.OnPaint()
    
    self.states["restore"] = ""

    self.LogMessage("Run Restore action ...")

    idx = int(self.vlblW2.GetFloatValue())
    scope = self.rsyncScope[idx]
    args = self.rsyncArgs[idx]
    if doDelete and args != "":
      args = f"{args} --del"
    
    restoredir = self.find_directory("vibes-backup", "/tmp/media")
    if not restoredir:
      self.LogMessage(f"Restore directory 'vibes-backup' not found ... stopping") 
      return

    cmd = f"rsync {args} --exclude '._*' -L --info=name1 {restoredir}{scope}/* /var/lib/emi/working{scope}"

    self.spinnerCtl(True)

    # Use the cmd variable as a single command string
    process = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    #process = subprocess.Popen(['python3', 'flashtest.py'], stdout=subprocess.PIPE, bufsize=256, universal_newlines=True)
    self.LogMessage("Start restore ...")

    # Read output while the process is running
    while process.poll() is None:
      
      output = process.stdout.read1().decode().strip()
      self.LogMessage(f"{output}")
    
    # Read remaining output after the process has finished
    #output = process.communicate()[0].decode().strip()
    #if output:
    #  self.LogMessage(output)

    self.LogMessage("Finalize ... ")
    self.runRepair()
    self.spinnerCtl(False)

    self.LogMessage("Restore ... done")


  def runReset(self):
    import tarfile
    import time
    
    def confirm_timer():
      import time
      time.sleep(5)  # Wait for 5s
      if self.states["facreset"] == "wait-to-confirm":
        self.LogMessage(f" Reset operation canceled. ")
        self.LogMessage(f" ") 
        self.states["facreset"] = ""
        btnW3=self.btnW3
        btnW3.SetText("Reset")
        btnW3.setRepaint()
        btnW3.OnPaint()

    if self.states["facreset"] == "":
      self.LogMessage(f" (!) Factory Reset will delete all user data ") 
      self.LogMessage(f" Do continue press wheelknob 3 or touch 'Yes' ") 
      self.LogMessage(f" Do cancel wait some seconds. ")
      self.states["facreset"] = "wait-to-confirm"
      self.confirm_thread = threading.Thread(target=confirm_timer)
      self.confirm_thread.start()
      btnW3=self.btnW3
      btnW3.SetText("Yes")
      btnW3.setRepaint()
      btnW3.OnPaint()
      return

    if self.states["facreset"] != "wait-to-confirm":
      return

    btnW3=self.btnW3
    btnW3.SetText("Reset")
    btnW3.setRepaint()
    btnW3.OnPaint()
    self.states["facreset"] = ""

    self.LogMessage("Run Reset action ...")
    self.spinnerCtl(True)
    try:
      factoryresetdatadir = self.find_directory("vibes-factory", "/tmp/media")

      if not factoryresetdatadir:
        self.LogMessage(f"Factory Reset directory 'vibes-factory' not found ... stopping") 
        self.spinnerCtl(False)
        return
      
      resetdatafile = os.path.join(factoryresetdatadir, "data.tar.gz")

      if not os.path.exists(resetdatafile):
        self.LogMessage(f"Factory Reset data file 'data.tar.gz' not found ... stopping")
        self.spinnerCtl(False)
        return

      self.LogMessage("Verify reset data ... ")
      if not self.check_files_hash(factoryresetdatadir, "resetdata.hash"):
          self.LogMessage("One or more hash checks failed ... stopping")
          self.spinnerCtl(False)
          return

      datadir = "/var/lib/emi"
      dataset = "default"
      
      # Forcefully delete all files and directories in datadir
      if os.path.exists(datadir):
          for item in os.listdir(datadir):
              item_path = os.path.join(datadir, item)
              if os.path.isdir(item_path):
                  shutil.rmtree(item_path)  # Remove directory and its contents
              else:
                  os.remove(item_path)  # Remove file

      extracttodir = os.path.join(datadir,dataset) 
      os.makedirs(extracttodir, exist_ok=True)
      self.LogMessage("Prepare reset data ... ")
      with tarfile.open(resetdatafile, 'r:gz') as tar:
          total_files = len(tar.getmembers())
          processed_files = 0

          # Function to show progress
          def show_progress():
              nonlocal processed_files
              while processed_files < total_files:
                  time.sleep(0.5)  # Wait for 500ms
                  self.LogMessage(f"{processed_files}/{total_files} items extracted.", True)

          # Start the progress display in a separate thread
          progress_thread = threading.Thread(target=show_progress)
          progress_thread.start()

          # Extract files
          for member in tar.getmembers():
              tar.extract(member, extracttodir)
              processed_files += 1

          # Wait for the progress thread to finish
          progress_thread.join()
          self.LogMessage(f"Factory items extracted ...")

      self.LogMessage(f"Create 'working' path ...")
      os.symlink(dataset, os.path.join(datadir, "working"))

      self.LogMessage(f"Adjust item attributes ...")
      self.recursive_chown(os.path.join(datadir, "working"), "emios", "emios")
      self.recursive_chmod(os.path.join(datadir, "working"), 0o770, 0o660)
    
    except Exception as e:
      self.LogMessage(f"ERROR: Failed: {e}")

    self.spinnerCtl(False)

  def runRepair(self):
    import fnmatch

    resfolder = "/var/lib/emi/working"

    for dirpath, dirnames, filenames in os.walk(resfolder):
        for filename in fnmatch.filter(filenames, '._*'):
            file_path = os.path.join(dirpath, filename)
            os.remove(file_path)  # Delete the file
            self.LogMessage(f"Deleted: {file_path}")

    self.LogMessage(f"Adjust item attributes ...")
    self.recursive_chown(resfolder, "emios", "emios")
    self.recursive_chmod(resfolder, 0o770, 0o660)

  def call_os_install(self, group, owner, perm, filename, installpath):
    # Call install tool 
    if os.path.exists(filename):
      result = subprocess.run(["install", "-g", group, "-o", owner, "-m", perm, filename, installpath], capture_output=True, text=True)
      self.LogMessage("'" + filename + "' installed to '" + installpath + "'")
    else:
      self.LogMessage("ERROR: '" + filename + "' not found.")

  def recursive_chown(self, path, user, group):
    for dirpath, dirnames, filenames in os.walk(path):
      shutil.chown(dirpath, user, group)
      for filename in filenames:
        if not os.path.islink(os.path.join(dirpath, filename)):
          shutil.chown(os.path.join(dirpath, filename), user, group)
          
  def recursive_chmod(self, path, dirperm, fileperm):
    for dirpath, dirnames, filenames in os.walk(path):
      os.chmod(dirpath, dirperm)
      for filename in filenames:
        if not os.path.islink(os.path.join(dirpath, filename)):
          os.chmod(os.path.join(dirpath, filename), fileperm)

  # Common 'Not Implemented' for CAN messages
  def NotImplemented(self, data, len):
    print(f"Not implemented")

  # Button / Label 'click' events
  def BtnOnEventHandler(self, event, btnid):
    btn = {
      1001: self.btnW1,
      1002: self.btnW2,
      1003: self.btnW3,
      # 1004: self.btnW4,
      9004: self.btnExit
    }
    if btnid in btn:
      btn[btnid].SetChecked(False)
      if btnid == 1001:
        self.runBackup()
      elif btnid == 1002:
        self.runRestore()
      elif btnid == 1003:
        self.runReset()
      # elif btnid == 1004:
      #  self.runTask004(3,4)
      elif btnid == 9004:
      #if btnid == 9004:
        self.OnExit()
